'''Cache netcdf data to disk to avoid re-computation.

Author: guangzhi XU (xugzhi1987@gmail.com)
Update time: 2020-07-16 21:31:35.
'''


#--------Import modules-------------------------
from __future__ import print_function
import sys, os
import functools
import cdms2 as cdms
import MV2 as MV
import numpy as np



def formFilename(folder, varid):
    '''Compose abspath for cache file

    Args:
        folder (str): cache folder path.
        varid (str): variable id to form file name and used as variable id.
    Returns:
        abpath (str): abspath for cache file, which is using the <folder>
            as folder. The file name is the format:
                [script_file]_[varid].nc
    '''
    script_file=os.path.splitext(sys.argv[0])[0]
    name='[%s]_[%s].nc' %(script_file, varid)
    abpath=os.path.join(folder, name)

    return abpath


def doRecomp(folder, varid_list, overwrite):
    '''Check whether to do recomputation or not

    Args:
        folder (str): cache folder path.
        varid_list (list): list of variable ids.
        overwrite (bool): whether to force a new computation or not.
    Returns:
        recmp (bool): if True, need to do recomputation, False otherwise.

    A check is performed on the existence of each file specified in
    <varid_list>, as `folder/file_name_for_varii`. `file_name_for_varii`
    is defined in formFilename(). If all files exist, no need to recompute,
    and return False. If any file missing, return True.

    If <overwrite> is True, return True to force a recomputation.
    '''

    recmp=False
    for vv in varid_list:
        abpath_in=formFilename(folder, vv)
        if not os.path.exists(abpath_in) or overwrite:
            recmp=True
            break

    return recmp


def forceList(arg):
    '''Make input argument into a list if not already so'''

    if isinstance(arg, (list, tuple)):
        arg_list=arg
    if isinstance(arg, str):
        arg_list=[arg, ]

    return arg_list


def readCache(folder, varid_list, overwrite=False, verbose=True):
    '''Read cached data

    Args:
        folder (str): cache folder path.
        varid_list (list): list of variable ids.
    Keyword Args:
        overwrite (bool): whether to force a new computation or not.
        verbose (bool): whether to print some text info.
    Returns:
        results (tuple): a tuple containing data read in from cached file(s).

    This func will form a file path for each variable given in <varid_list>,
    and save the read data into a tuple.
    '''

    results=[]
    for vv in varid_list:
        #---------------Form data filename---------------
        abpath_in=formFilename(folder, vv)
        if os.path.exists(abpath_in) and not overwrite:

            if verbose:
                print('\n# <readCache>: Read in variable', vv,
                        'from disk cache:\n', abpath_in)

            fin=cdms.open(abpath_in,'r')
            vii=fin(vv)
            fin.close()
            results.append(vii)

    return tuple(results)


def writeCache(results, folder, varid_list, verbose=True):
    '''Write data to disk cache

    Args:
        results (tuple): a tuple containing data read to cache.
        folder (str): cache folder path.
        varid_list (list): list of variable ids.
    Keyword Args:
        verbose (bool): whether to print some text info.

    This func will form a file path for each variable given in <varid_list>,
    and save the read data into a cach file.
    '''

    for varii, idii in zip(results, varid_list):
        abpath_in=formFilename(folder, idii)
        fout=cdms.open(abpath_in,'w')
        if verbose:
            print('\n# <writeCache>: Saving output to:\n',abpath_in)
        varii.id=idii
        fout.write(varii,typecode='f')
        fout.close()

    return


class DiskCacher(object):
    def __init__(self, varid, func=None, folder=None, overwrite=False,
            verbose=True):
        '''Disk cache context manager

        Args:
            varid (str or list or tuple): if str, function <func> is assumed
                to return only 1 return value. If list or tuple, each variable
                id corresponds to one return value of <func>.
        Keyword Args:
            func (callable): function object whose return values are to be
                cached.
            folder (str or None): cache folder path. If None, use a default.
            overwrite (bool): whether to force a new computation or not.
            verbose (bool): whether to print some text info.
        '''

        if folder is None:
            self.folder='/home/guangzhi/datasets/quicksave2/'
        else:
            self.folder=folder

        self.func=func
        self.varid=varid
        self.overwrite=overwrite
        self.verbose=verbose

        #---------------Put varids to a list---------------
        self.varid_list=forceList(varid)

    def __enter__(self):
        '''Called in the `with` statement, return a decorated function
        '''
        if self.func is None:
            raise Exception("Need to provide a callable function to __init__() when used as context manager.")

        return _Cache2Disk(self.func, self.varid_list, self.folder,
                self.overwrite, self.verbose)

    def __exit__(self, type, value, traceback):
        '''Nothing to do after `with`'''
        return

    def __call__(self, func=None):
        _func=func or self.func
        return _Cache2Disk(_func, self.varid_list, self.folder,
                self.overwrite, self.verbose)


"""
def Cache2Disk(varid, func=None, folder=None, overwrite=False, verbose=True):
    '''Disk cacher decorator function

    Args:
        varid (str or list or tuple): if str, function <func> is assumed
            to return only 1 return value. If list or tuple, each variable
            id corresponds to one return value of <func>.
    Keyword Args:
        func (callable or None): function object whose return values are to be
            cached. If None, to be used as a decorator:

                ```
                @Cache2Disk(['varid1', 'varid2'])
                def myFunc(arg1, arg2, ...):
                    ...
                ```

            If callable, can be used as:
                ```
                newfunc = Cache2Disk(['varid1', varid2'], myFunc)
                var1, var2 = newfunc(arg1, arg2, ...)
                ```
        folder (str or None): cache folder path. If None, use a default.
        overwrite (bool): whether to force a new computation or not.
        verbose (bool): whether to print some text info.

    NOTE that the modified function can return one variable, or multiple,
    but they have to be all netcdf data that can be saved by cdms.
    '''

    if folder is None:
        folder='/home/guangzhi/datasets/quicksave2/'
    else:
        folder=folder

    #---------------Put varids to a list---------------
    varid_list=forceList(varid)

    if func is None:
        return functools.partial(_Cache2Disk, varid_list=varid_list,
                folder=folder, overwrite=overwrite, verbose=verbose)
    else:
        return _Cache2Disk(func, varid_list, folder, overwrite, verbose)
"""


def _Cache2Disk(func, varid_list, folder, overwrite, verbose):
    '''Inner decorator function

    Args:
        func (callable): function object whose return values are to be
            cached.
        varid_list (list): list of variable ids.
        folder (str): cache folder path.
        overwrite (bool): whether to force a new computation or not.
        verbose (bool): whether to print some text info.
    Returns:
        decorated function: if cache exists, the function is <readCache>
            which will read cached data from disk. If needs to recompute,
            the function is wrapped that the return values are saved to disk
            before returning.
    '''

    def decorator_func(func):

        @functools.wraps(func)
        def wrapper(*args, **kwargs):
            if not doRecomp(folder, varid_list, overwrite):
                #----------Cache exists, do no recompute----------
                results=readCache(folder, varid_list, overwrite, verbose)
            else:
                #----------Cache not exists, do recompute----------
                results=func(*args, **kwargs)
                if isinstance(results, tuple):
                    if len(results)!=len(varid_list):
                        raise Exception("Length of result doesn't equal varid.")
                else:
                    results=(results,)
                if not os.path.exists(folder):
                    os.makedirs(folder)
                writeCache(results, folder, varid_list, verbose)

            if len(results)==1:
                return results[0]
            else:
                return results
            return results
        return wrapper

    return decorator_func(func)


def createDummy():

    lats=np.arange(0, 90)
    lons=np.arange(0, 360)
    xx,yy=np.meshgrid(lons, lats)
    z=2*np.sin(xx*np.pi/180) + 3*np.cos(2*yy*np.pi/180)
    z=MV.array(z)

    latax=cdms.createAxis(lats)
    latax.designateLatitude()
    latax.id='y'
    latax.units='degree north'
    latax.name='latitude'

    lonax=cdms.createAxis(lons)
    lonax.designateLongitude()
    lonax.id='x'
    lonax.units='degree east'
    lonax.name='longitude'

    z.setAxisList([latax, lonax])
    z.id='z'
    z.units='1'
    z.long_name='dummy'

    return z




#-------------Main---------------------------------
if __name__=='__main__':

    #--------------Create some dummy data--------------
    var=createDummy()


    #-------------Usage 1: context manager-------------
    print('\n\n#########################################')
    print('Usage in context manager')

    def func1(var, n, lat1=10, lat2=30):
        '''dummy test function'''

        result=var[slice(1,3)]*n
        result2=var(latitude=(lat1, lat2))*n*2
        print('### Computing in function.\n')

        return result, result2

    with DiskCacher(['context_a', 'context_b'], func1) as func1b:
        result=func1b(var, 10, lat1=10, lat2=30)

    with DiskCacher(['context_a', 'context_b'], func1) as func1b:
        result=func1b(var, 10, lat1=10, lat2=30)

    with DiskCacher(['context_a', 'context_b'], func1, overwrite=True)\
            as func1b:
        result=func1b(var, 10, lat1=10, lat2=30)

    print('#########################################')

    #-------------Usage 2: wrap a function-------------
    print('\n\n#########################################')
    print('Usage as function wrap')

    def func2(var, n, lat1=10, lat2=30):
        '''dummy test function'''

        result=var[slice(1,3)]*n
        result2=var(latitude=(lat1, lat2))*n*2
        print('### Computing in function.\n')

        return result, result2

    func2b=DiskCacher(['wrap_a', 'wrap_b'])(func2)
    result=func2b(var, 10, lat1=10, lat2=30)

    func2b=DiskCacher(['wrap_a', 'wrap_b'])(func2)
    result=func2b(var, 10, lat1=10, lat2=30)

    func2b=DiskCacher(['wrap_a', 'wrap_b'], overwrite=True)(func2)
    result=func2b(var, 10, lat1=10, lat2=30)

    print('#########################################')



    #-------------Usage 3: decorate a function-------------
    print('\n\n#########################################')
    print('Usage as function decorator')

    @DiskCacher(['dec_a', 'dec_b'])
    def func3(var, n, lat1=10, lat2=30):
        '''dummy test function'''

        result=var[slice(1,3)]*n
        result2=var(latitude=(lat1, lat2))*n*2
        print('### Computing in function.\n')

        return result, result2

    result=func3(var, 10, lat1=10, lat2=30)
    result=func3(var, 10, lat1=10, lat2=30)

    print('#########################################')









